import psycopg2
import subprocess
import os
import pandas as pd
from psycopg2.extensions import connection as Connection, cursor as Cursor
from psycopg2.extras import execute_values
from logging import Logger
from typing import List, Dict, Any, Optional
from stpstone.cals.handling_dates import DatesBR
from stpstone.handling_data.json import JsonFiles
from stpstone.loggs.create_logs import CreateLog


class PostgreSQLDB:

    def __init__(self, dbname:str, user:str, password:str, host:str, port:int, 
                 str_schema:str='public', logger:Optional[Logger]=None) \
        -> None:
        '''
        DOCSTRING: INITIALIZES THE CONNECTION TO THE POSTGRESQL DATABASE
        INPUTS: 
            - DBNAME:STR
            - USER:STR
            - PASSWORD:STR
            - HOST:STR
            - PORT:INT
            - SCHEMA:STR
            - LOGGER:OPTIONAL[LOGGER]
        OUTPUTS: -
        '''
        self.dbname = dbname
        self.user = user
        self.password = password
        self.host = host
        self.port = port
        self.str_schema = str_schema
        self.logger = logger
        self.dict_db_config = {
            'dbname': self.dbname,
            'user': self.user,
            'password': self.password,
            'host': self.host,
            'port': self.port
        }
        self.conn: Connection = psycopg2.connect(**self.dict_db_config)
        self.cursor: Cursor = self.conn.cursor()
        self._execute(f"SET search_path TO '{self.str_schema}';")

    def _execute(self, str_query:str) -> None:
        '''
        DOCSTRING: RUN QUERY WITH DML ACCESS
        INPUTS: QUERY
        OUTPUTS: -
        '''
        self.cursor.execute(str_query)

    def _read(self, str_query:str, dict_type_cols:Optional[Dict[str, Any]]=None, 
              list_cols_dt:Optional[List[str]]=None, str_fmt_dt:Optional[str]=None) -> pd.DataFrame:
        '''
        DOCSTRING: RUN QUERY AND RETURN RESULTS AS DATAFRAME
        INPUTS: QUERY, DICT_TYPE_COLS, LIST_COLS_DT, STR_FMT_DT
        OUTPUTS: Pandas DataFrame
        '''
        # retrieving dataframe
        df_ = pd.read_sql_query(str_query, self.conn)
        # changing data types
        if all([x is not None for x in [dict_type_cols, list_cols_dt, str_fmt_dt]]):
            df_ = df_.astype(dict_type_cols)
            for col_ in list_cols_dt:
                df_[col_] = [DatesBR().str_date_to_datetime(d, str_fmt_dt) for d in df_[col_]]
        # return dataframe
        return df_

    def _insert(self, json_data:List[Dict[str, Any]], str_table_name:str, 
        bl_insert_or_ignore:bool=False) -> None:
        '''
        DOCSTRING: INSERTS DATA FROM A JSON OBJECT INTO A POSTGRESQL TABLE
        INPUTS: JSON_DATA
        OUTPUTS: -
        '''
        # validate json, in order to have the same keys
        json_data = JsonFiles().normalize_json_keys(json_data)
        # sql insert statement
        list_columns = ', '.join(json_data[0].keys())
        list_placeholders = ', '.join(['%s' for _ in json_data[0]])
        if bl_insert_or_ignore == True:
            str_query = f'INSERT INTO {str_table_name} ({list_columns}) VALUES ' \
                + f'({list_placeholders}) ON CONFLICT DO NOTHING'
        else:
            str_query = f'INSERT INTO {str_table_name} ({list_columns}) VALUES ' \
                + f'({list_placeholders})'
        try:
            execute_values(
                self.cursor,
                str_query,
                [tuple(record.values()) for record in json_data]
            )
            self.conn.commit()
            if self.logger is not None:
                CreateLog().infos(
                    self.logger, 
                    f'Successful commit in db {self.dict_db_config["dbname"]} ' 
                    + f'/ table {str_table_name}.'
                )
        except Exception as e:
            self.conn.rollback()
            self._close
            if self.logger is not None:
                CreateLog().errors(
                    self.logger, 
                    'ERROR WHILE INSERTING DATA\n'
                    + f'DB_CONFIG: {self.dict_db_config}\n'
                    + f'TABLE_NAME: {str_table_name}\n'
                    + f'JSON_DATA: {json_data}\n'
                    + f'ERROR_MESSAGE: {e}'
                )
            raise Exception(
                'ERROR WHILE INSERTING DATA\n'
                + f'DB_CONFIG: {self.dict_db_config}\n'
                + f'TABLE_NAME: {str_table_name}\n'
                + f'JSON_DATA: {json_data}\n'
                + f'ERROR_MESSAGE: {e}'
            )

    @property
    def _close(self) -> None:
        '''
        DOCSTRING: CLOSES THE CONNECTION TO THE DATABASE
        INPUTS: -
        OUTPUTS: -
        '''
        self.conn.close()
    
    def _bkp_db(self, str_backup_dir:str, str_bkp_name:str=None) -> str:
        '''
        DOCSTRING: BACKUP DATABASE - NEED PG_DUMP TO BE CONFIGURED WITHIN ENVIRONMENT VARIABLES
        INPUTS: 
            - BACKUP_DIR (STR): DIRECTORY WHERE THE BACKUP FILE WILL BE SAVED
        OUTPUTS: STR: PATH TO THE BACKUP FILE, OR AN ERROR MESSAGE.
        '''
        try:
            # ensure the backup directory exists
            os.makedirs(str_backup_dir, exist_ok=True)
            # generate the backup file name
            backup_file = os.path.join(str_backup_dir, str_bkp_name)
            # set the environment variable for self.password authentication
            env = os.environ.copy()
            env['PGPASSWORD'] = self.password
            # run the pg_dump command
            command = [
                'pg_dump',
                '-h', self.host,
                '-p', str(self.port),
                '-U', self.user,
                '-F', 'c',  # custom format
                '-b',       # include large objects
                '-f', backup_file,  # output file
                self.dbname
            ]
            subprocess.run(command, check=True, env=env)
            return f'Backup successful! File saved at: {backup_file}'
        except subprocess.CalledProcessError as e:
            return f'Backup failed: {e}'
        except Exception as e:
            return f'An error occurred: {e}'


### CHARTS FOR RISK MANAGEMENT ###

import matplotlib.pyplot as plt
from stpstone.finance.financial_risk.market_risk import MarketRiskManagement


class ChartsRskManagement:

    def efficient_frontier_makowitz(self, list_stocks_weights_portfolio, list_stocks_tickers,
                                    date_inf, date_sup, risk_free_nominal_rate,
                                    num_portfolios=25000, num_working_days_year=252,
                                    country='brazil'):
        '''
        REFERENCES: https://towardsdatascience.com/efficient-frontier-portfolio-optimisation-in-python-e7844051e7f
        DOCSTRING: CHART OF 
        INPUTS:
        OUTPUTS:
        '''
        # setup for markowitz portfolio
        setup_markowitz_portfolio = \
            MarketRiskManagement().setup_markowitz_portfolio(list_stocks_weights_portfolio,
                                                             list_stocks_tickers,
                                                             date_inf, date_sup,
                                                             num_working_days_year, country)

        # json n markowitz portfolios
        json_markowitz_random_portfolios = \
            MarketRiskManagement().efficient_frontier_portfolio(
                setup_markowitz_portfolio['array_daily_stocks_returns'],
                setup_markowitz_portfolio['array_mean_daily_stocks_returns'],
                risk_free_nominal_rate, num_portfolios, num_working_days_year)

        # arguments for plot: std, return and sharpe ratio, respectively
        x_axis_std = [d['markowitz_portfolio']['std_portfolio_nominal']
                      for d in json_markowitz_random_portfolios['n_potfolios'].values()]
        y_axis_returns = [d['markowitz_portfolio']['return_portfolio_stocks_nominal']
                          for d in json_markowitz_random_portfolios['n_potfolios'].values()]
        colorify_trigger = [d['markowitz_portfolio']['sharpe_ratio']
                            for d in json_markowitz_random_portfolios['n_potfolios'].values()]

        # plot - std x return
        plt.figure(figsize=(10, 7))
        plt.scatter(x_axis_std, y_axis_returns, c=colorify_trigger, cmap='YlGnBu', marker='o', s=10,
                    alpha=0.3)
        plt.colorbar()

        # maximum sharpe ratio
        plt.scatter(json_markowitz_random_portfolios['max_sharpe_ratio_portfolio']['std_nominal'],
                    json_markowitz_random_portfolios['max_sharpe_ratio_portfolio']['return_nominal'],
                    marker='*', color='r', s=500, label='Maximum Sharpe Ratio')

        # minimum volatility
        plt.scatter(json_markowitz_random_portfolios['min_volatility_portfolio']['std_nominal'],
                    json_markowitz_random_portfolios['min_volatility_portfolio']['return_nominal'],
                    marker='*', color='g', s=500, label='Minimum Volatility')

        # chart elements
        plt.title('Simulated Portfolio Optimization Based on Efficient Frontier')
        plt.xlabel('Annualised Volatility')
        plt.ylabel('Annualised Returns')
        plt.legend(labelspacing=0.8)
        plt.show()


# ChartsRskManagement().efficient_frontier_makowitz([0.3, 0.25, 0.45],
#                                                   ['PETR3', 'ITSA4', 'WEGE3'],
#                                                   '01/01/2016', '04/09/2020', 0.02, 100000)

# ChartsRskManagement().efficient_frontier_makowitz([0.25, 0.25, 0.25, 0.25],
#                                                   ['AAPL', 'AMZN', 'GOOGL', 'FB'],
#                                                   '01/01/2016', '31/12/2017', 0.02, 100000,
#                                                   country='United States')



### MODULE TO HANDLE DATABRICKSCLI INTEGRATION ###

import pandas as pd
from time import sleep
import os
import json
import datetime


class JobsCLI:

    def __init__(self, job_id, path='C:\Temp'):
        self.job_id = job_id
        self.path = path

    def print_time(self, complete_path):
        command = r'echo %date%-%time% > {}'.format(complete_path)
        stream = os.popen(command)

    def job_run(self, filename="run_id.json", notebook_params=None):
        complete_path = os.path.join(self.path, filename)
        if notebook_params:
            command = """databricks jobs run-now --job-id {} -- notebook-params '{}'> {}""".format(
                self.job_id, str(notebook_params).replace("'", '"'), complete_path)
        else:
            command = "databricks jobs run-now --job-id {} > {}".format(
                self.job_id, complete_path)
        stream = os.popen(command)
        sleep(5)
        with open(complete_path) as f:
            dict_metadata = json.load(f)
        self.run_id = dict_metadata['run_id']

    def get_job_metadata(self, filename="job_metadata.json"):
        complete_path = os.path.join(self.path, filename)
        command = 'databricks jobs get --job-id {} > {}'.format(
            self.job_id, complete_path)
        stream = os.popen(command)
        sleep(5)
        with open(complete_path) as f:
            dict_metadata = json.load(f)
        return dict_metadata

    def cancel_job_run(self, outside_run_id=None):
        if outside_run_id:
            command = 'databricks runs cancel --run-id {}'.format(
                outside_run_id)
        else:
            command = 'databricks runs cancel --run-id {}'.format(self.run_id)
        stream = os.popen(command)

    def get_run_metadata(self, filename="run_metadata.json", outside_run_id=None):
        complete_path = os.path.join(self.path, filename)
        if outside_run_id:
            command = 'databricks runs get --run-id {} > {}'.format(
                outside_run_id, complete_path)
        else:
            command = 'databricks runs get --run-id {} > {}'.format(
                self.run_id, complete_path)
        stream = os.popen(command)
        sleep(5)
        with open(complete_path) as f:
            dict_metadata = json.load(f)
        return dict_metadata

    def get_run_output(self, filename="run_output_metadata.json", outside_run_id=None):
        complete_path = os.path.join(self.path, filename)
        if outside_run_id:
            command = 'databricks runs get-output --run-id {} > {}'.format(outside_run_id,
                                                                           complete_path)
        else:
            command = 'databricks runs get-output --run-id {} > {}'.format(self.run_id,
                                                                           complete_path)
        stream = os.popen(command)
        sleep(5)
        with open(complete_path) as f:
            dict_metadata = json.load(f)
        return dict_metadata


class DbfsCLI:

    def copy(self, path_ori, path_dest, overwrite=True):
        if overwrite:
            command = 'dbfs cp "{}" "{}" --overwrite'.format(
                path_ori, path_dest)
        else:
            command = 'dbfs cp "{}" "{}"'.format(path_ori, path_dest)
        print(command)
        stream = os.popen(command)
        output = [x.strip() for x in stream.readlines()]

    def remove(self, path):
        command = 'dbfs rm "{}"'.format(path)
        stream = os.popen(command)
        output = [x.strip() for x in stream.readlines()]
        return output[1]

    def move(self, path_ori, path_dest):
        command = 'dbfs mv "{}" "{}"'.format(path_ori, path_dest)
        stream = os.popen(command)

    def list_files(self, path, absolute='--absolute', l='-l'):
        command = 'dbfs ls "{}" {} {}'.format(path, absolute, l)
        stream = os.popen(command)
        if l == '-l':
            output = [x.strip().split(' ') for x in stream.readlines()]
            for x in output:
                while '' in x:
                    x.remove('')
            dict_infos = {x[2]: {'tipo': x[0], 'tamanho': x[1]}
                          for x in output}
            return dict_infos
        else:
            output = [x.strip() for x in stream.readlines()]
            return output

    def copy_and_run(self, local_path, dbfs_path, job_id, int_seconds_wait=10):
        '''
        DOCSTRING: COPY CSV FILES TO DBFS IN DATABRICKS WORKSPACE THAN RUN THE JOB TO UPLOAD THE DATA
        INPUTS: LOCAL PATH, DBFS PATH AND THE JOB ID IN DATABRICKS
        OUTPUTS: -
        '''
        self.copy(local_path, dbfs_path)
        tables = self.list_files('dbfs:/FileStore/tables')
        while dbfs_path not in tables.keys():
            self.copy(local_path, dbfs_path)
            tables = self.list_files('dbfs:/FileStore/tables')
        sleep(int_seconds_wait)
        db_cli = JobsCLI(job_id)
        db_cli.job_run()
        sleep(int_seconds_wait)
        try:
            run_state = db_cli.get_run_metadata()
        except json.JSONDecodeError:
            print('Unable to fetch metadata from run: {}'.format(db_cli.run_id))
        sleep(int_seconds_wait)
        while run_state['state']['life_cycle_state'] != 'TERMINATED':
            try:
                run_state = db_cli.get_run_metadata()
            except json.JSONDecodeError:
                print('Unable to fetch metadata from run: {}'.format(db_cli.run_id))
            sleep(int_seconds_wait)
        # status of accomplishment
        if run_state['state']['result_state'] != 'SUCCESS':
            return 'Error running job'
        else:
            return 'Success running job'



### GROGRAPHICAL INFO WORLWIDE ###

from requests import request
import ast
import sys
sys.path.append(r'C:\Users\Guilherme\OneDrive\Dev\Python\Packages')
from stpstone.handling_data.str import StrHandler


class BrazilGeo:

    def states(self, bol_accents=True, key_nome='nome'):
        '''
        DOCSTRING: RETURN BRAZILLIAN STATES
        INPUTS: BOOLEAN ACCENTS (WHETER SHOULD BE INCLUDED IN FEDERATIVE UNITY OR NOT, 
            BY DEFAULT TRUE)
        OUTPUTS: SERIALIZED JSON WITH STATES ID, SHORT NAME AND REGION
        '''
        # fetching states of brazillian territory
        url_localidades_brasil = 'https://servicodados.ibge.gov.br/api/v1/localidades/estados'
        # rest consult to ibge database
        response = request('GET', url_localidades_brasil)
        dict_message = ast.literal_eval(StrHandler().find_between(
            str(response.text.encode('utf8')), "b'", "'"))
        # latin characters to brazillian states names
        # wheter remove or not accents according to user will
        for dict_ativo in dict_message:
            dict_ativo[key_nome] = StrHandler().latin_characters(
                dict_ativo[key_nome])
            if bol_accents == False:
                dict_ativo[key_nome] = StrHandler(
                ).removing_accents(dict_ativo[key_nome])
        # sending response
        return dict_message

    def zip_code(self, list_zip_codes, url='https://cep.awesomeapi.com.br/json/{}',
                 method='GET', key_address_type='adress_type',
                 key_address='address', key_state='state'):
        '''
        DOCSTRING: ZIP CODE LOCATION INFO
        INPUTS: LIST ZIP CODES, URL (DEFAULT), METHOD (DEFAULT)
        OUTPUTS: DICTIONARY
        '''
        # setting variables
        dict_zip_adresses = dict()
        # looping through each zip code, requesting info about adrees, and raise exception
        #   if status code is different from 2xx
        for zip_ in list_zip_codes:
            resp_req = request(method=method, url=url.format(zip_))
            #   raises exception when not a 2xx response
            resp_req.raise_for_status()
            #   generating json
            json_zip_codes = resp_req.json()
            #   organizing data to export format
            dict_zip_adresses[zip_] = [json_zip_codes[key_address_type],
                                       json_zip_codes[key_address], json_zip_codes[key_state]]
        # return data with zip code location info
        return dict_zip_adresses


# print(BrazilGeo().states(bol_accents=False))
# # output
# [{'id': 11, 'sigla': 'RO', 'nome': 'Rondonia', 'regiao': {'id': 1, 'sigla': 'N', 'nome': 'Norte'}}, {'id': 12, 'sigla': 'AC', 'nome': 'Acre', 'regiao': {'id': 1, 'sigla': 'N', 'nome': 'Norte'}}, {'id': 13, 'sigla': 'AM', 'nome': 'Amazonas', 'regiao': {'id': 1, 'sigla': 'N', 'nome': 'Norte'}}, {'id': 14, 'sigla': 'RR', 'nome': 'Roraima', 'regiao': {'id': 1, 'sigla': 'N', 'nome': 'Norte'}}, {'id': 15, 'sigla': 'PA', 'nome': 'Para', 'regiao': {'id': 1, 'sigla': 'N', 'nome': 'Norte'}}, {'id': 16, 'sigla': 'AP', 'nome': 'Amapa', 'regiao': {'id': 1, 'sigla': 'N', 'nome': 'Norte'}}, {'id': 17, 'sigla': 'TO', 'nome': 'Tocantins', 'regiao': {'id': 1, 'sigla': 'N', 'nome': 'Norte'}}, {'id': 21, 'sigla': 'MA', 'nome': 'Maranhao', 'regiao': {'id': 2, 'sigla': 'NE', 'nome': 'Nordeste'}}, {'id': 22, 'sigla': 'PI', 'nome': 'Piaui', 'regiao': {'id': 2, 'sigla': 'NE', 'nome': 'Nordeste'}}, {'id': 23, 'sigla': 'CE', 'nome': 'Ceara', 'regiao': {'id': 2, 'sigla': 'NE', 'nome': 'Nordeste'}}, {'id': 24, 'sigla': 'RN', 'nome': 'Rio Grande do Norte', 'regiao': {'id': 2, 'sigla': 'NE', 'nome': 'Nordeste'}}, {'id': 25, 'sigla': 'PB', 'nome': 'Paraiba', 'regiao': {'id': 2, 'sigla': 'NE', 'nome': 'Nordeste'}}, {'id': 26, 'sigla': 'PE', 'nome': 'Pernambuco', 'regiao': {'id': 2, 'sigla': 'NE', 'nome': 'Nordeste'}}, {'id': 27, 'sigla': 'AL', 'nome': 'Alagoas', 'regiao': {'id': 2, 'sigla': 'NE', 'nome': 'Nordeste'}}, {'id': 28, 'sigla': 'SE', 'nome': 'Sergipe', 'regiao': {'id': 2, 'sigla': 'NE', 'nome': 'Nordeste'}}, {'id': 29, 'sigla': 'BA', 'nome': 'Bahia', 'regiao': {'id': 2, 'sigla': 'NE', 'nome': 'Nordeste'}}, {'id': 31, 'sigla': 'MG', 'nome': 'Minas Gerais', 'regiao': {'id': 3, 'sigla': 'SE', 'nome': 'Sudeste'}}, {'id': 32, 'sigla': 'ES', 'nome': 'Espirito Santo', 'regiao': {'id': 3, 'sigla': 'SE', 'nome': 'Sudeste'}}, {'id': 33, 'sigla': 'RJ', 'nome': 'Rio de Janeiro', 'regiao': {'id': 3, 'sigla': 'SE', 'nome': 'Sudeste'}}, {'id': 35, 'sigla': 'SP', 'nome': 'Sao Paulo', 'regiao': {'id': 3, 'sigla': 'SE', 'nome': 'Sudeste'}}, {'id': 41, 'sigla': 'PR', 'nome': 'Parana', 'regiao': {'id': 4, 'sigla': 'S', 'nome': 'Sul'}}, {'id': 42, 'sigla': 'SC', 'nome': 'Santa Catarina', 'regiao': {'id': 4, 'sigla': 'S', 'nome': 'Sul'}}, {'id': 43, 'sigla': 'RS', 'nome': 'Rio Grande do Sul', 'regiao': {'id': 4, 'sigla': 'S', 'nome': 'Sul'}}, {'id': 50, 'sigla': 'MS', 'nome': 'Mato Grosso do Sul', 'regiao': {'id': 5, 'sigla': 'CO', 'nome': 'Centro-Oeste'}}, {'id': 51, 'sigla': 'MT', 'nome': 'Mato Grosso', 'regiao': {'id': 5, 'sigla': 'CO', 'nome': 'Centro-Oeste'}}, {'id': 52, 'sigla': 'GO', 'nome': 'Goias', 'regiao': {'id': 5, 'sigla': 'CO', 'nome': 'Centro-Oeste'}}, {'id': 53, 'sigla': 'DF', 'nome': 'Distrito Federal', 'regiao': {'id': 5, 'sigla': 'CO', 'nome': 'Centro-Oeste'}}]




### DEAL WITH PANDAS ISSUES ###

import os
import pandas as pd
from typing import List, Dict, Any
from stpstone.settings._global_slots import YAML_MICROSOFT_APPS
from stpstone.cals.handling_dates import DatesBR
from stpstone.handling_data.folders import DirFilesManagement
from stpstone.loggs.create_logs import CreateLog
from stpstone.microsoft_apps.excel import DealingExcel


class DealingPd(DealingExcel):
    '''
    DOCSTRING: COMMON FUNCTIONS TO DEAL WITH DATA IMPORT TO PANDAS DATAFRAMES OR TIMESERIES
    '''

    def append_df_to_Excel(self, filename, list_tup_df_sheet_name, bl_header=True, bl_index=0, 
        mode='w', label_sensitivity='internal', bl_set_sensitivity_label=False, bl_debug_mode=False):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # debug mode
        if bl_debug_mode == True:
            print('LIST_TUP_DF_SHEET_NAME: {}'.format(list_tup_df_sheet_name))
            print('FILENAME: {}'.format(filename))
            print('MODE: {}'.format(mode))
            print('BL_INDEX: {}'.format(bl_index))
        # write excel
        with pd.ExcelWriter(filename, engine='xlsxwriter', mode=mode) as writer:
            for df_, sheet_name in list_tup_df_sheet_name:
                # # checking df_
                # if df_.empty == True: continue
                # removing indexes
                if bl_index == 0:
                    df_.reset_index(drop=True, inplace=True)
                    # df_ = df_.style.hide_index()
                df_.to_excel(writer, sheet_name, index=bl_index, header=bl_header)
        # setting label
        if bl_set_sensitivity_label == True:
            self.xlsx_sensitivity_label(filename, YAML_MICROSOFT_APPS[
                'sensitivity_labels_office'], 
                label_sensitivity.capitalize())

    def export_xl(self, logger, nome_completo_xlsx_exportacao, list_tup_df_sheet_name, 
        range_colunas='A:CC', bl_adjust_layout=False, bl_debug_mode=False):
        '''
        DOCSTRING: EXPORTANDO DAFRAME PARA EXCEL
        INPUTS: LOGGER, NOME COMPLETO XLSX DE EXPOTAÇÃO, DATAFRAME DE EXPORTAÇÃO, NOME DA PLANILHA, 
            RANGE DE COLUNAS, BOOLEAN AUTOAJUSTE (DEFAULT)
        OUTPUTS: BOOLEAN
        '''
        #   exportando dataframe para excel
        DealingPd().append_df_to_Excel(
            nome_completo_xlsx_exportacao, list_tup_df_sheet_name, bl_header=True, bl_index=0, 
            bl_debug_mode=bl_debug_mode)
        #   validando se o arquivo foi exportado com sucesso
        blame_exportacao_xlsx = DirFilesManagement().object_exists(
            nome_completo_xlsx_exportacao)
        if blame_exportacao_xlsx == 'NOK':
            CreateLog().warnings(logger, 'Arquivo não salvo na rede: {}'.format(
                nome_completo_xlsx_exportacao))
            raise Exception('Arquivo {} não salvo na rede, favor validar'.format(
                nome_completo_xlsx_exportacao))
        else:
            if bl_adjust_layout == True:
                for _, plan_nome in list_tup_df_sheet_name:
                    #   corrigindo formato do xlsx exportado, e aplicando um autofit nas colunas 
                    #       de interesse
                    xla, wb = self.open_xl(nome_completo_xlsx_exportacao)
                    self.autofit_range_columns(plan_nome, range_colunas, xla, wb)
                    self.close_wb(wb)
        return blame_exportacao_xlsx

    def json_to_excel(self, json_path_name, xlsx_path_name):
        '''
        DOCSTRING: EXPORT JSON FILE TO EXCEL XLSX
        INPUTS: JSON AND XLSX COMPLETE NAME
        OUTPUTS: STATUS OF ACCOMPLISHMENT
        '''
        pd.read_json(json_path_name).to_excel(xlsx_path_name)
        if os.path.exists(xlsx_path_name):
            return 'OK'
        else:
            return 'NOK'

    def json_normalizer(self, json_path):
        '''
        DOCSTRING: JSON NORMALIZER FROM PANDAS
        INPUTS: JSON PATH
        OUTPUTS: JSON IN DATAFRAME
        '''
        return pd.read_json(json_path)

    def settingup_pandas(self):
        '''
        DOCSTRING: CONFIGURAÇÃO BÁSICA DO PANDAS
        INPUTS: COLUNA COMO STR
        OUTPUTS: -
        '''
        # Use 3 decimal places in output display
        pd.set_option("display.precision", 3)
        # Don't wrap repr(DataFrame) across additional lines
        pd.set_option("display.expand_frame_repr", False)
        # Set max rows displayed in output to 25
        pd.set_option("display.max_rows", 25)

    def dataframe_to_dict(self, dataframe_export, orientation='records'):
        '''
        REFERENCES: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_dict.html
        DOCSTRING: EXPORT DATAFRAME TO JSON
        INPUTS: DATAFRAME TO BE EXPORTED, ORIENTATION(SPLIT, RECORDS (DEFAULT), INDEX, COLUMNS, VALUES, 
            TABLE)
        OUTPUTS: JSON
        '''
        return dataframe_export.to_dict(orient=orientation)

    def convert_datetime_columns(self, df_, list_col_date, bl_pandas_convertion=True):
        '''
        DOCSTRING: CONVERTING DATE COLUMNS TO DATETIME
        INPUTS: DATAFRAME, COLUMNS WITH DATE INFO
        OUTPUTS: DATAFRAME
        '''
        # checking wheter to covert through a pandas convertion, or resort to a excel format transformation
        #   of data in date column format
        if bl_pandas_convertion:
            for col_date in list_col_date:
                df_.loc[:, col_date] = pd.to_datetime(
                    df_[col_date], unit='s').dt.date
        else:
            # corventing list column dates to string type
            df_ = df_.astype(
                dict(zip(list_col_date, [str] * len(list_col_date))))
            # looping through each row
            for index, row in df_.iterrows():
                for col_date in list_col_date:
                    if '-' in row[col_date]:
                        ano = int(row[col_date].split(' ')[0].split('-')[0])
                        mes = int(row[col_date].split(' ')[0].split('-')[1])
                        dia = int(row[col_date].split(' ')[0].split('-')[2])
                        df_.loc[index, col_date] = DatesBR(
                        ).build_date(ano, mes, dia)
                    else:
                        df_.loc[index, col_date] = DatesBR(
                        ).excel_float_to_date(int(row[col_date]))
        # returning dataframe with date column to datetime forma
        return df_

    def merge_dfs_into_df(self, df_1, df_2, cols_list):
        '''
        DOCSTRING: MERGING TWO DFS INTO A NEW ONE, REMOVING THEIR INTERSECTION ON A LIST OF COLUMNS
        INPUTS: TWO DATAFRAMES AND A LIST WITH COLUMNS NAMES
        OUTPUTS: DATAFRAME
        '''
        intersec_df = pd.merge(df_1, df_2, how='inner', on=cols_list)
        merge_df = pd.merge(df_2, intersec_df, how='outer',
                            on=cols_list, indicator=True)
        merge_df = merge_df.loc[merge_df._merge == 'left_only']
        merge_df.dropna(how='all', axis=1, inplace=True)
        try:
            merge_df = merge_df.drop(columns='_merge')
        except:
            pass
        for column in merge_df.columns:
            if column in cols_list:
                pass
            else:
                merge_df = merge_df.rename(
                    columns={str(column): str(column)[:-2]})
        return merge_df

    def max_chrs_per_column(df_:pd.DataFrame) -> dict:
        '''
        DOCSTRING: CALCULATE THE MAXIMUM NUMBER OF CHARACTERS IN A SPECIFIED COLUMN OF A DATAFRAME
        INPUTS: DATAFRAME
        OUTPUTS: DICT
        '''
        dict_ = dict()
        for col_ in list(df_.columns):
            dict_[col_] = df_[col_].astype(str).str.len().max()
        return dict_

    def change_dtypes(self, df_:pd.DataFrame, dict_dtypes:Dict[str, Any], list_cols_dt:List[str], 
        str_fmt_dt:str='YYYY-MM-DD', errors:str='raise') -> pd.DataFrame:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        if any([col_ not in list(dict_dtypes.keys()) for col_ in list_cols_dt]):
            for col_ in list(dict_dtypes.keys()):
                dict_dtypes[col_] = str
        df_ = df_.astype(dict_dtypes, errors=errors)
        for col_ in list_cols_dt:
            df_[col_] = [DatesBR().str_date_to_datetime(d, str_fmt_dt) for d in df_[col_]]
        return df_

    def strip_all_obj_dtypes(self, df_:pd.DataFrame, list_cols_dt:List[str]) -> pd.DataFrame:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        list_cols = [col_ for col_ in list(df_.select_dtypes(['object']).columns) if col_ 
            not in list_cols_dt]
        for col_ in list_cols:
            df_[col_] = [x.strip() for x in df_[col_]]
        return df_

    def fillna_data(self, df_:pd.DataFrame, list_cols_dt:List[str], str_dt_fillna:str='2100-12-31', 
        str_data_fillna:str='-1') -> pd.DataFrame:
        '''
        DOCSTRING: FILLNA DATA/DATES
        INPUTS: DATAFRAME, LIST COLS DT, LIST COLS DATA
        OUTPUTS: DATAFRAME
        '''
        list_cols = [col_ for col_ in list(df_.columns) if col_ not in list_cols_dt]
        if list_cols_dt is not None:
            df_[list_cols_dt] = df_[list_cols_dt].fillna(str_dt_fillna)
        df_[list_cols] = df_[list_cols].fillna(str_data_fillna)
        return df_

    def pipeline_df_startup(self, df_:pd.DataFrame, dict_dtypes:Dict[str, Any], 
        list_cols_dt:List[str], str_dt_fillna:str='2100-12-31', str_data_fillna:str='-1', 
        str_fmt_dt:str='YYYY-MM-DD'):
        '''
        DOCSTRING: PIPELINE DATAFRAME STARTUP
        INPUTS: DATAFRAME, DICT DTYPES, LIST OF DATE TYPE COLUMNS, STR DT FILLNA, STR DATA FILLNA, 
            STR DATE FORMAT INPUT
        OUTPUTS:
        '''
        df_ = self.fillna_data(df_, list_cols_dt, str_dt_fillna, str_data_fillna)
        df_ = self.change_dtypes(df_, dict_dtypes, list_cols_dt, str_fmt_dt)
        df_ = self.strip_all_obj_dtypes(df_, list_cols_dt)
        return df_

    def cols_remove_dupl(self, df_):
        '''
        DOCSTRING: REMOVE DUPLICATES WITHIN COLUMNS AND KEEP JUST THE FIRST ONE
        INPUTS: DATAFRAME
        OUTPUTS: DATAFRAME
        '''
        return df_.loc[:, ~df_.columns.duplicated()]



### PROBABILITY DISTRIBUTIONS ###

from __future__ import print_function, division
import numpy as np
import matplotlib.pylab as plt
import seaborn as sns
from scipy.special import gamma, gammaln
from scipy.stats import t, uniform, chi2, f, bernoulli, geom, binom, poisson, sem, norm
from numpy import shape, dot, ones, multiply, pi, sqrt, log


class ProbabilityDistributions:

    def bernoulli_distribution(self, prob, num_trials=1):
        '''
        REFERENCES: https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.stats.bernoulli.html
        DOCSTRING: BERNOULLI DISTRIBUTION TO ANALYZE PERCENTAGE OF ACCOMPLISHMENT AND FAILURE
            FOR EACH EVENT: P(x=1) = P, E(X) = P, V(X) = P * (1-P)
        INPUTS: PROBABILITY AND NUMBER OF TRIALS
        OUTPUTS: ARRAY OF BERNOULLI DISTRIBUTION (MEAN, VAR, SKEW, KURT AND
            CUMULATIVE DISTRIBUTION FUNCTION)
        '''
        return {
            'mean': bernoulli.stats(prob, moments='mvsk')[0],
            'var': bernoulli.stats(prob, moments='mvsk')[1],
            'skew': bernoulli.stats(prob, moments='mvsk')[2],
            'kurt': bernoulli.stats(prob, moments='mvsk')[3],
            'distribution': bernoulli.cdf(num_trials, prob)
        }

    def geometric_distribution(self, prob, num_trials):
        '''
        REFERENCES: http://biorpy.blogspot.com/2015/02/py19-geometric-distribution-in-python.html
        DOCSTRING: GEOMETRIC DISTRIBUTION TO INDICATE NUMBER OF INDEPENDENT TRIALS TO REACH
            FIRST SUCCESS: P(X=N) = (1-P) ** (N-1) * P, E(X) = 1/P, V(X) = (1-P) / P ** 2
        INPUTS: PROBABILITY (FLOAT) AND NUMBER OF TRIALS
        OUTPUTS: DICT OF GEOMETRIC DISTRIBUTION (MEAN, VAR, SKEW, KURT AND
            CUMULATIVE DISTRIBUTION FUNCTION)
        '''
        p = np.zeros(num_trials)
        for k in range(1, num_trials + 1):
            p[k - 1] = geom.pmf(k, prob)
        return {
            'mean': geom.stats(p, moments='mvsk')[0],
            'var': geom.stats(p, moments='mvsk')[1],
            'skew': geom.stats(p, moments='mvsk')[2],
            'kurt': geom.stats(p, moments='mvsk')[3],
            'distribution': p
        }

    def binomial_distribution(self, prob, num_trials):
        '''
        REFERENCES: https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.stats.binom.html
        DOCSTRING: BINOMIAL DISRTIBUTION TO INVESTIGATE K-NUMBER OF SUCCESSES IN N-TRIALS:
            P(X=K) = COMB(N,K) * P ** K * (1 - P) ** (N-K), E(X) = N * P, V (X) = N * P * (1-P)
        INPUTS: PROBABILITY (FLOAT) AND NUMBER OF TRIALS
        OUTPUTS: DICT OF BINOMIAL DISTRIBUTION (MEAN, VAR, SKEW, KURT AND
            CUMULATIVE DISTRIBUTION FUNCTION)
        '''
        p = np.zeros(num_trials)
        for k in range(1, num_trials + 1):
            p[k - 1] = binom.pmf(k, num_trials, prob)
        return {
            'mean': binom.stats(num_trials, p, moments='mvsk')[0],
            'var': binom.stats(num_trials, p, moments='mvsk')[1],
            'skew': binom.stats(num_trials, p, moments='mvsk')[2],
            'kurt': binom.stats(num_trials, p, moments='mvsk')[3],
            'distribution': p
        }

    def poisson_distribution(self, num_trials, mu):
        '''
        REFERENCES: https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.poisson.html
        DOCSTRING: POISSON DISTRIBUTION TO COUNT OCCURRENCES WITHIN CERTAIN AMOUNT OF TIME, ASSUMING
            A LAMBDA MEAN:
            P(X=K) = EXP(-LAMBDA) * LAMBA ** K / K!, E(x) = V(X) = LAMBDA * RANGE
        INPUTS: NUMBER OF TRIALS AND MU
        OUTPUTS: DICT OF POISSON DISTRIBUTION (MEAN, VAR, SKEW, KURT AND
            CUMULATIVE DISTRIBUTION FUNCTION)
        '''
        p = np.zeros(num_trials)
        for k in range(1, num_trials + 1):
            p[k - 1] = poisson.pmf(k, mu)
        return {
            'mean': poisson.stats(mu, moments='mvsk')[0],
            'var': poisson.stats(mu, moments='mvsk')[1],
            'skew': poisson.stats(mu, moments='mvsk')[2],
            'kurt': poisson.stats(mu, moments='mvsk')[3],
            'distribution': p
        }

    def chi_squared(self, p, df, probability_func='ppf', x_axis_inf_range=None,
                    x_axis_sup_range=None, x_axis_pace=None):
        '''
        DOCSTRING: CHI SQUARED PROBABILITY POINT FUNCTION (Z-SCORE, OR PPF), PROBABABILITY 
            DENSITY FUNCTION (PDF), AND PROBABILITY CUMULATIVE FUNCTION (CDF)
        INPUTS: P (PROBABILITY), DEGREES OF FREEDOM, PROBABILITY FUNCTION (PPF, AS DEFAULT, WHEREAS 
            PDF AND CDF ARE POSSIBLE AS WELL)
        OUTPUTS: FLOAT
        '''
        # setting x axis range
        arr_ind = np.arange(x_axis_inf_range, x_axis_sup_range, x_axis_pace)
        # getting the statistic
        if probability_func == 'ppf':
            return chi2.ppf(p, df)
        elif probability_func == 'pdf':
            return chi2.pdf(arr_ind, df)
        elif probability_func == 'cdf':
            return chi2.cdf(arr_ind, df)
        else:
            raise Exception('Error defining the probability function of interest. {} '.format(
                probability_func)
                + 'was given, nevertheless was expected ppf, pdf or cdf')

    def t_student(self, p, df, probability_func='ppf', x_axis_inf_range=None,
                  x_axis_sup_range=None, x_axis_pace=None):
        ''' 
        DOCSTRING: T STUDENT PROBABILITY POINT FUNCTION (Z-SCORE, OR PPF), PROBABABILITY 
            DENSITY FUNCTION (PDF), AND PROBABILITY CUMULATIVE FUNCTION (CDF)
        INPUTS: P (PROBABILITY), DEGREES OF FREEDOM, PROBABILITY FUNCTION (PPF, AS DEFAULT, WHEREAS 
            PDF AND CDF ARE POSSIBLE AS WELL)
        OUTPUTS: FLOAT
        '''
        # setting x axis range
        arr_ind = np.arange(x_axis_inf_range, x_axis_sup_range, x_axis_pace)
        # getting the statistic
        if probability_func == 'ppf':
            return t.ppf(p, df)
        elif probability_func == 'pdf':
            return t.pdf(arr_ind, df)
        elif probability_func == 'cdf':
            return t.cdf(arr_ind, df)
        else:
            raise Exception('Error defining the probability function of interest. {} '.format(
                probability_func)
                + 'was given, nevertheless was expected ppf, pdf or cdf')

    def f_fisher_snedecor(self, dfn, dfd, mu, p=None, probability_func='ppf', x_axis_inf_range=None,
                          x_axis_sup_range=None, x_axis_pace=None):
        '''
        DOCSTRING: F-SNEDECOR PROBABILITY POINT FUNCTION (Z-SCORE, OR PPF), PROBABABILITY 
            DENSITY FUNCTION (PDF), AND PROBABILITY CUMULATIVE FUNCTION (CDF)
        INPUTS: P (PROBABILITY), DEGREES OF FREEDOM, PROBABILITY FUNCTION (PPF, AS DEFAULT, WHEREAS 
            PDF AND CDF ARE POSSIBLE AS WELL)
        OUTPUTS: FLOAT
        '''
        # checking wheter degrees of freedom numerator is higher than denominator
        assert dfn > dfd
        # defining basic parameters of the distribution
        f_dist = f(dfn, dfd, mu)
        # setting x axis range
        arr_ind = np.arange(x_axis_inf_range, x_axis_sup_range, x_axis_pace)
        # getting the statistic
        if probability_func == 'ppf':
            return f.ppf(p, dfn, dfd)
        elif probability_func == 'pdf':
            return f_dist.pdf(arr_ind)
        elif probability_func == 'cdf':
            return f.cdf(p, dfn, dfd)
        else:
            raise Exception('Error defining the probability function of interest. {} '.format(
                probability_func)
                + 'was given, nevertheless was expected ppf, pdf or cdf')


class NormalDistribution:

    def phi(self, x):
        '''
        DOCSTRING: RETURN THE VALUE OF THE GAUSSIAN PROBABILITY FUNCTION WITH MEAN 0.0 AND
            STANDARD DEVIATION 1.0 AT THE GIVEN X VALUE
        INPUTS: X
        OUTPUS: STANDARD NORMAL PROBABILITY
        '''
        return np.exp(-x ** 2 / 2.0) / np.sqrt(2.0 * np.pi)

    def pdf(self, x, mu=0.0, sigma=1.0):
        '''
        DOCSTRING: RETURN THE VALUE OF THE GAUSSIAN PROBABILITY FUNCTION WITH MEAN MU AND
            STANDARD DEVIATION SIGMA AT THE GIVEN x VALUE
        INPUTS: X, MU (0.0 BY DEFAULT) AND SIGMA (1.0 BY DEFAULT)
        OUTPUTS: VALUE OF THE GAUSSIAN PROBABILITY
        '''
        return self.phi((x - mu) / sigma) / sigma

    def cumnulative_phi(self, z):
        '''
        DOCSTRING: DENSITY FUNCTION WITH MEAN 0.0 AND STANDARD DEVIATION 1.0 AT THE GIVEN Z VALUE
        INPUTS: Z
        OUTPUTS: PHI
        '''
        if z < -8.0:
            return 0.0
        if z > 8.0:
            return 1.0
        total = 0.0
        term = z
        i = 3
        while total != total + term:
            total += term
            term *= z * z / float(i)
            i += 2
        return 0.5 + total * self.phi(z)

    def cdf(self, x, mu=0.0, sigma=1.0):
        '''
        DOCSTRING: STANDARD GAUSSIAN CDF WITH MEAN MI AND STDDEV SIGMA, USING TAYLOR
            APPROXIMATION - CUMULATIVE DISTRIBUTION FUNCTION - AREA BELOW GAUSSIAN CURVE -
            NORMAL DISTRIBUTION FORMULA
        INPUTS: X, MU(STANDARD 0.0) AND SIGMA (STANDARD 1.0)
        OUTPUTS: CUMULATIVE DENSITY FUNCTION OF A GAUSSIAN DISTRIBUTION
        '''
        return self.cumnulative_phi((x - mu) / sigma)

    def inv_cdf(self, p, mu=0.0, sigma=1.0):
        '''
        DOCSTRING: INVERSE OF THE NORMAL CULMULATIVE DISTRIBUTION FOR A SUPPLIED VALUE OF X, OR
            A PROBABILITY, WITH A GIVEN DISTRIBUTION MEAND AND STANDARD DEVIATION
        INPUTS: PROBABILITY, MEAN AND STANDARD DEVIATION
        OUTPUTS: INV.NORM, OR Z-SCORE
        '''
        return norm.ppf(p, mu, sigma)

    def confidence_interval_normal(self, data, confidence=0.95):
        '''
        REFERENCE: https://stackoverflow.com/questions/15033511/compute-a-confidence-interval-from-sample-data
        DOCSTRING: CONFIDECENCE INTERVAL FOR A NORMAL DISTRIBUTION
        INPUTS: DATA AND CONFIDENCE
        OUTPUTS: DICTIONARY (MEAN, INFERIOR AND SUPERIOR INTERVALS)
        '''
        a = 1.0 * np.array(data)
        n = len(a)
        mu, se = np.mean(a), sem(a)
        z = se * t.ppf((1 + confidence) / 2., n - 1)
        return {
            'mean': mu,
            'inferior_inteval': mu - z,
            'superior_interval': mu + z
        }

    def ecdf(self, data):
        '''
        REFERENCES: https://campus.datacamp.com/courses/statistical-thinking-in-python-part-1/graphical-exploratory-data-analysis?ex=12
        DOCSTRING: COMPUTE ECDF FOR A ONE-DIMENSIONAL ARRAY OF MEASUREMENTS AN EMPIRICAL 
            CUMULATIVE DISTRIBUTION FUNCTION (ECDF)
        INPUTS: DATA
        OUTPUTS: X-AXIS AND Y-AXIS
        '''
        # number of data points: n
        n = len(data)
        # x-data for the ECDF: x
        x = np.sort(data)
        # y-data for the ECDF: y
        y = np.arange(1, n + 1) / n
        return x, y


class HansenSkewStudent(object):

    '''Skewed Student distribution class - This is the version introduced by Bruce E. Hansen in 1994.
    References: https://www.ssc.wisc.edu/~bhansen/papers/ier_94.pdf
    Attributes
    ----------
    eta : float
        Degrees of freedom. :math:`2 < \eta < \infty`
    lam : float
        Skewness. :math:`-1 < \lambda < 1`
    Methods
    -------
    pdf
        Probability density function (PDF)
    cdf
        Cumulative density function (CDF)
    ppf
        Inverse cumulative density function (ICDF)
    rvs
        Random variates with mean zero and unit variance
    '''

    def __init__(self, eta=10., lam=-.1):
        '''Initialize the class.
        Parameters
        ----------
        eta : float
            Degrees of freedom. :math:`2 < \eta < \infty`
        lam : float
            Skewness. :math:`-1 < \lambda < 1`
        '''
        self.eta = eta
        self.lam = lam

    @property
    def const_a(self):
        '''Compute a constant.
        Returns
        -------
        a : float
        '''
        return 4 * self.lam * self.const_c() * (self.eta - 2) / (self.eta - 1)

    @property
    def const_b(self):
        '''Compute b constant.
        Returns
        -------
        b : float
        '''
        return (1 + 3 * self.lam**2 - self.const_a**2)**.5

    def const_c(self):
        '''Compute c constant.
        Returns
        -------
        c : float
        '''
        return gamma((self.eta + 1) / 2) \
            / ((np.pi * (self.eta - 2))**.5 * gamma(self.eta / 2))

    def pdf(self, arg):
        '''Probability density function (PDF).
        Parameters
        ----------
        arg : array
            Grid of point to evaluate PDF at
        Returns
        -------
        array
            PDF values. Same shape as the input.
        '''
        c = self.const_c()
        a = self.const_a
        b = self.const_b

        return b * c * (1 + 1 / (self.eta - 2) * ((b * arg + a) / (1 + np.sign(
            arg + a / b) * self.lam))**2)**(-(self.eta + 1) / 2)

    def cdf(self, arg):
        '''Cumulative density function (CDF).
        Parameters
        ----------
        arg : array
            Grid of point to evaluate CDF at
        Returns
        -------
        array
            CDF values. Same shape as the input.
        '''
        a = self.const_a
        b = self.const_b

        y = (b * arg + a) / (1 + np.sign(arg + a / b) * self.lam) * (
            1 - 2 / self.eta)**(-.5)
        cond = arg < -a / b

        return cond * (1 - self.lam) * t.cdf(y, self.eta) \
            + ~cond * (-self.lam + (1 + self.lam) * t.cdf(y, self.eta))

    def ppf(self, arg):
        '''Inverse cumulative density function (ICDF).
        Parameters
        ----------
        arg : array
            Grid of point to evaluate ICDF at. Must belong to (0, 1)
        Returns
        -------
        array
            ICDF values. Same shape as the input.
        '''
        arg = np.atleast_1d(arg)

        a = self.const_a
        b = self.const_b

        cond = arg < (1 - self.lam) / 2

        ppf1 = t.ppf(arg / (1 - self.lam), self.eta)
        ppf2 = t.ppf(.5 + (arg - (1 - self.lam) / 2) /
                     (1 + self.lam), self.eta)
        ppf = -999.99 * np.ones_like(arg)
        ppf = np.nan_to_num(ppf1) * cond \
            + np.nan_to_num(ppf2) * np.logical_not(cond)
        ppf = (ppf * (1 + np.sign(arg - (1 - self.lam) / 2) * self.lam) * (
            1 - 2 / self.eta)**.5 - a) / b

        if ppf.shape == (1, ):
            return float(ppf)
        else:
            return ppf

    def rvs(self, size=1):
        '''Random variates with mean zero and unit variance.
        Parameters
        ----------
        size : int or tuple
            Size of output array
        Returns
        -------
        array
            Array of random variates
        '''
        return self.ppf(uniform.rvs(size=size))

    def plot_pdf(self, arg=np.linspace(-2, 2, 100)):
        '''Plot probability density function.
        Parameters
        ----------
        arg : array
            Grid of point to evaluate PDF at
        '''
        scale = (self.eta / (self.eta - 2))**.5
        plt.plot(arg, t.pdf(arg, self.eta, scale=1 / scale),
                 label='t distribution')
        plt.plot(arg, self.pdf(arg), label='skew-t distribution')
        plt.legend()
        plt.show()

    def plot_cdf(self, arg=np.linspace(-2, 2, 100)):
        '''Plot cumulative density function.
        Parameters
        ----------
        arg : array
            Grid of point to evaluate CDF at
        '''
        scale = (self.eta / (self.eta - 2))**.5
        plt.plot(arg, t.cdf(arg, self.eta, scale=1 / scale),
                 label='t distribution')
        plt.plot(arg, self.cdf(arg), label='skew-t distribution')
        plt.legend()
        plt.show()

    def plot_ppf(self, arg=np.linspace(.01, .99, 100)):
        '''Plot inverse cumulative density function.
        Parameters
        ----------
        arg : array
            Grid of point to evaluate ICDF at
        '''
        scale = (self.eta / (self.eta - 2))**.5
        plt.plot(arg, t.ppf(arg, self.eta, scale=1 / scale),
                 label='t distribution')
        plt.plot(arg, self.ppf(arg), label='skew-t distribution')
        plt.legend()
        plt.show()

    def plot_rvspdf(self, arg=np.linspace(-2, 2, 100), size=1000):
        '''Plot kernel density estimate of a random sample.
        Parameters
        ----------
        arg : array
            Grid of point to evaluate ICDF at. Must belong to (0, 1)
        '''
        rvs = self.rvs(size=size)
        xrange = [arg.min(), arg.max()]
        sns.kdeplot(rvs, clip=xrange, label='kernel')
        plt.plot(arg, self.pdf(arg), label='true pdf')
        plt.xlim(xrange)
        plt.legend()
        plt.show()

    def loglikelihood(theta=None, x=None):
        nu = theta[0]

        lambda_ = theta[1]

        c = gamma((nu + 1) / 2) / \
            (multiply(sqrt(dot(pi, (nu - 2))), gamma(nu / 2)))

        a = multiply(multiply(dot(4, lambda_), c), ((nu - 2) / (nu - 1)))

        b = sqrt(1 + dot(3, lambda_ ** 2) - a ** 2)

        logc = gammaln((nu + 1) / 2) - gammaln(nu / 2) - \
            dot(0.5, log(dot(pi, (nu - 2))))

        logb = dot(0.5, log(1 + dot(3, lambda_ ** 2) - a ** 2))

        find1 = (x < (- a / b))

        find2 = (x >= (- a / b))

        LL1 = logb + logc - dot((nu + 1) / 2.0, log(1 + multiply(1.0 / (nu - 2), ((
            multiply(b, x) + a) / (1 - lambda_)) ** 2)))

        LL2 = logb + logc - dot((nu + 1) / 2.0, log(1 + multiply(1.0 / (nu - 2), ((
            multiply(b, x) + a) / (1 + lambda_)) ** 2)))

        LL = sum(LL1[find1]) + sum(LL2[find2])

        LL = -LL

        return LL.sum()



### FUNÇÕES ÚTEIS PARA INTERACAO ENTRE DATAS ###

import locale
import time
import businesstimedelta
import pytz
import pandas as pd
from datetime import timedelta, date, datetime, time, timezone
from dateutil.relativedelta import relativedelta
from workalendar.core import MON, SAT, SUN
from more_itertools import unique_everseen
from pprint import pprint
from stpstone.cals.br_bzdays import BrazilBankCalendar
from stpstone.handling_data.str import StrHandler


class DatesBR(BrazilBankCalendar):
    '''
    DOCSTING: CALENDÁRIO DE DIAS ÚTEIS POR CIDADES NO BRASIL, CALENDÁRIO DE FERIADOS BANCÁRIOS E
    FUNÇÕES COM DATAS (ÚLTIMO DIA ÚTIL BANCÁRIO, PRÓXIMO DIA ÚTIL, É DIA ÚTIL?, CONVERSÃO DE DATAS
    DE STR PARA INT (VISANDO REALIZAR CÁLCULOS COM A DATA) E INTERVALO DE DATAS ÚTEIS ENTRE OS DOIS
    LIMITES (PASSADOS COMO ARGUMENTO)
    '''

    def build_date(self, year, month, day):
        '''
        DOCSTRING: BUILD DATETIME WITH YEAR, MONTH AND DAY INFO
        INPUTS: YEAR, MONTH, DAY
        OUTPUTS: DATETIME
        '''
        return date(year=year, month=month, day=day)

    def build_datetime(self, year, month, day, hour, minute, second):
        '''
        DOCSTRING: BUILD DATETIME WITH YEAR, MONTH AND DAY INFO
        INPUTS: YEAR, MONTH, DAY
        OUTPUTS: DATETIME
        '''
        return datetime(year=year, month=month, day=day, hour=hour, minute=minute, second=second)

    def date_to_datetime(self, date, bl_crop_time=True, bl_tinestamp=True):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        datetime_ = datetime.combine(date, datetime.min.time())
        if bl_tinestamp == True:
            datetime_ = datetime_.timestamp()
        if bl_crop_time == True:
            return int(datetime_)
        else:
            return datetime_

    def to_integer(self, dt_time):
        '''
        DOCSTRING: CONVERT DATETIME TO INTEGER
        INPUTS: DATETIME
        OUTPUTS: INTEGER
        '''
        return 10000 * dt_time.year + 100 * dt_time.month + dt_time.day

    def excel_float_to_date(self, float_excel_date):
        '''
        DOCSTRING: RETRIEVE DATE FROM EXCEL DATE FORMAT
        INPUTS: EXCEL DATE
        OUTPUTS: DATE
        '''
        return date.fromordinal(date(1900, 1, 1).toordinal() + float_excel_date - 2)

    def excel_float_to_datetime(self, float_excel_date):
        '''
        DOCSTRING: RETRIEVE DATETIME FROM EXCEL DATE FORMAT
        INPUTS: EXCEL DATE
        OUTPUTS: DATETIME
        '''
        return datetime.fromordinal(datetime(1900, 1, 1).toordinal() + float_excel_date - 2)

    def check_date_datetime_format(self, curr_date):
        '''
        DOCSTRING: CHECK WHETER OR NOT THE DATE FORMAT IS DATETIME
        INPUTS: DATE
        OUTPUTS: BOOLEAN
        '''
        return isinstance(curr_date, date)

    def str_date_to_datetime(self, data_str, format='DD/MM/YYYY'):
        '''
        DOCSTRING: CONVERSÃO DE DATA DE STRING PARA DATETIME.DATE, VISANDO REALIZAR CONTAS COM A MESMA
        INPUTS: DATA STRING NO format DD/MM/YYYY, YYYY-MM-DD, YYMMDD, DDMMYY OU DDMMYYYY
        OUTPUTS: DATA INT NO format YYYY-MM-DD
        '''
        if format == 'DD/MM/YYYY':
            return date(int(data_str[-4:]), int(data_str[3:5]), int(data_str[0:2]))
        elif format == 'YYYY-MM-DD':
            return date(int(data_str[0:4]), int(data_str[5:7]), int(data_str[-2:]))
        elif format == 'YYMMDD':
            return date(int('20' + data_str[0:2]), int(data_str[2:4]), int(data_str[-2:]))
        elif format == 'DDMMYY':
            return date(int('20' + data_str[-2:]), int(data_str[2:4]), int(data_str[0:2]))
        elif format == 'DDMMYYYY':
            return date(int(data_str[0:2]), int(data_str[2:4]), int(data_str[-4:]))
        elif format == 'DD/MM/YY':
            int_year = int(str(DatesBR().year_number(DatesBR().curr_date))[:2] + str(data_str[-2:]))
            return date(int_year, int(data_str[3:5]), int(data_str[0:2]))
        else:
            raise Exception('format não é válido, favor revisite o parâmetro')

    def datetime_to_string(self, curr_date, date_format='%d/%m/%Y'):
        '''
        DOCSTRING: DATETIME TO STRING IN DESIGNED FORMAT
        INPUTS: CURRENTE DATE IN DATETIME AND DATE FORMAT
        OUTPUTS: STRING
        '''
        return curr_date.strftime(date_format)

    def date_to_datetime(self, date, bl_crop_time=True, bl_tinestamp=True):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        datetime_ = datetime.combine(date, datetime.min.time())
        if bl_tinestamp == True:
            datetime_ = datetime_.timestamp()
        if bl_crop_time == True:
            return int(datetime_)
        else:
            return datetime_

    def list_range_number_working_days_between_dates(self, data_inic, data_fim, du_ant_ref_date=0,
                                                     format_datas_input=None):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # correção format datas
        if type(data_inic) == str:
            data_inic = DatesBR().str_date_to_datetime(data_inic, format_datas_input)
        if type(data_fim) == str:
            data_fim = DatesBR().str_date_to_datetime(data_fim, format_datas_input)
        # criando data de referência
        ref_date = self.sub_working_days(self.curr_date(), du_ant_ref_date)
        # definindo du_anteriores inf e sup, em relação à data de interesse
        du_inf = self.get_working_days_delta(data_inic, ref_date)
        du_sup = self.get_working_days_delta(data_fim, ref_date)
        # retornando range de du anteriores
        return range(du_sup, du_inf + 1)

    def list_working_days(self, data_inic, data_fim, format_datas_input=None, format_datas_saida=None):
        '''
        DOCSTRING: DIAS ÚTEIS ENTRE DUAS DATAS, CONSIDERANDO OS FERIADOS BANCÁRIOS NO BRASIL
        INPUTS: DATA DE INÍCIO, DATA FIM EM DATETIME.DATE, STR ('DD/MM/YYYY' OU 'YYYY-MM-DD')
            E format DA DATA
        OUTPUTS: LISTA DE DIAS ÚTEIS ENTRE DUAS DATAS EM DATETIME.DATE
        '''
        # correção format datas
        if str(type(data_inic)) == "<class 'str'>":
            data_inic = DatesBR().str_date_to_datetime(data_inic, format_datas_input)
        if str(type(data_fim)) == "<class 'str'>":
            data_fim = DatesBR().str_date_to_datetime(data_fim, format_datas_input)
        # lista de passagem
        list_du = list()
        # * super() just works in its own scope, however list comprehensions create its own scope
        # * so we have to append an empty list
        for x in range(int((data_fim - data_inic).days) + 1):
            list_du.append(super().find_following_working_day(
                day=data_inic + timedelta(days=x)))
        if format_datas_saida != None:
            return[DatesBR().datetime_to_string(d, format_datas_saida) for d in
                   list(unique_everseen(list_du))]
        else:
            return list(unique_everseen(list_du))

    def list_calendar_days(self, data_inic, data_fim, format_data='DD/MM/YYYY'):
        '''
        DOCSTRING: DIAS CORRIDOS ENTRE DUAS DATAS
        INPUTS: DATA DE INÍCIO E DATA FIM EM DATETIME.DATE (INT)
        OUTPUTS: LISTA DE DIAS ÚTEIS ENTRE DUAS DATAS EM DATETIME.DATE
        '''
        # correção format datas
        if str(type(data_inic)) == "<class 'str'>":
            data_inic = DatesBR().str_date_to_datetime(data_inic, format_data)
        if str(type(data_fim)) == "<class 'str'>":
            data_fim = DatesBR().str_date_to_datetime(data_fim, format_data)
        # lista de passagem
        list_du = list()
        # appendando dias úteis
        # * super() just works in its own scope, however list comprehensions create its own scope
        # * so we have to append an empty list
        for x in range(int((data_fim - data_inic).days)):
            list_du.append(data_inic + timedelta(days=x))
        return list(unique_everseen(list_du))

    def list_years_within_dates(self, data_inic, data_fim):
        '''
        DOCSTRING: LISTA DE ANOS ENTRE DATAS
        INPUTS: DATA DE INÍCIO E DATA FIM EM DATETIME.DATE (INT)
        OUTPUTS: LISTA DE ANOS ENTRE DUAS DATAS
        '''
        # correção format datas
        if str(type(data_inic)) == "<class 'str'>":
            data_inic = DatesBR().str_date_to_datetime(data_inic, 'DD/MM/YYYY')
        if str(type(data_fim)) == "<class 'str'>":
            data_fim = DatesBR().str_date_to_datetime(data_fim, 'DD/MM/YYYY')
        # lista de passagem
        list_years = list()
        # appendando dias úteis
        # * super() just works in its own scope, however list comprehensions create its own scope
        # * so we have to append an empty list
        for x in range(int((data_fim - data_inic).days)):
            list_years.append((data_inic + timedelta(days=x)).year)
        return list(unique_everseen(list_years))

    @property
    def curr_date(self):
        '''
        DOCSTRING: RETORNA A DATA DE HOJE EM DATETIME.DATE
        INPUTS: -
        OUTPUTS: DATA DE HOJE EM DATETIME.DATE
        '''
        return date.today()

    @property
    def curr_time(self):
        '''
        DOCSTRING: RETORNA O HORÁRIO LOCAL NO MOMENTO DA EXECUÇÃO DA DEF
        INPUTS: -
        OUTPUTS: HORA DE AGORA EM DATETIME.DATE
        '''
        return datetime.now().time()

    def curr_date_time(self, bl_timestamp=False, bl_crop_time=False):
        '''
        DOCSTRING: RETORNA HORA E DATA LOCAL ATUAL
        INPUTS: -
        OUTPUTS: DATA E HORA DE AGORA NO format EM DATETIME.DATE
        '''
        # validating wheter the user wants to return the timestamp format or not
        if bl_timestamp == True:
            datetime_ = datetime.now().timestamp()
        else:
            datetime_ = datetime.now()
        # crop time if its user's will
        if bl_crop_time == True:
            return int(datetime_)
        else:
            return datetime_

    def testing_dates(self, data_inic, data_fim):
        '''
        DOCSTRING: TESTAR SE DATAS INSERIDAS RESPEIAM UM CRITÉRIO LÓGICO
        INPUTS: DATA DE INÍCIO E DE FIM
        OUTPUTS: RESPOSTA À SOLICITAÇÃO
        '''
        if type(data_inic) != datetime:
            data_inic = DatesBR().date_to_datetime(data_inic)
        if type(data_fim) != datetime:
            data_fim = DatesBR().date_to_datetime(data_fim)
        if int((data_fim - data_inic).days) >= 0:
            return True
        else:
            return True

    def year_number(self, date):
        '''
        DOCSTRING: YEAR NUMBER
        INPUTS: DATE (AS DATETIME)
        OUPUTS: INTEGER
        '''
        return int(date.strftime('%Y'))

    def day_number(self, date):
        '''
        DOCSTRING: MONTH NUMBER
        INPUTS: DATE (AS DATETIME)
        OUPUTS: INTEGER
        '''
        return int(date.strftime('%d'))

    def month_name(self, date, abbreviation=False, local_zone='pt-BR'):
        '''
        DOCSTRING: NAME OF THE MONTH IN THE LOCAL LANGUAGE
        INPUTS: DATE, ABBREVIATION AND LOCAL ZONE (BRAZIL AS STANDARD)
        OUTPUTS: STRING WITH THE NAME OF THE MONTH
        '''
        locale.setlocale(locale.LC_TIME, local_zone)
        if abbreviation == True:
            return date.strftime('%b')
        else:
            return date.strftime('%B')

    def date_begin_end_month(self, date_, last_month_year=12, format='YYYY-MM-DD'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # converting input date to the correct format
        if type(date_) == str:
            self.str_date_to_datetime(date_, format)
        # defining inputs for the dates range
        year = self.year_number(date_)
        month = self.month_number(date_)
        day = 1
        # building inferior and superior dates
        date_inf = self.find_working_day(self.build_date(year, month, day))
        if month < last_month_year:
            date_sup = self.sub_working_days(
                self.build_date(year, month + 1, day), 1)
        else:
            date_sup = self.sub_working_days(
                self.build_date(year + 1, 1, day), 1)
        # returning dates
        return date_inf, date_sup

    def month_number(self, date, bl_month_mm=False):
        '''
        DOCSTRING: MONTH NUMBER
        INPUTS: DATE (AS DATETIME)
        OUPUTS: INTEGER/STRING
        '''
        if bl_month_mm == False:
            return int(date.strftime('%m'))
        else:
            return date.strftime('%m')

    def week_name(self, date, abbreviation=False, local_zone='pt-BR'):
        '''
        DOCSTRING: NAME OF THE WEEKDAY IN THE LOCAL LANGUAGE
        INPUTS: DATE, ABBREVIATION AND LOCAL ZONE (BRAZIL AS STANDARD)
        OUTPUTS: STRING WITH THE NAME OF THE WEEK
        '''
        locale.setlocale(locale.LC_TIME, local_zone)
        if abbreviation == True:
            return date.strftime('%a')
        else:
            return date.strftime('%A')

    def week_number(self, date):
        '''
        DOCSTRING: NUMBER OF THE WEEKDAY
        INPUTS: DATE (AS DATETIME)
        OUTPUTS: INT
        '''
        return date.strftime('%w')

    def find_working_day(self, date_test):
        '''
        DOCSTRING: FIND NEXT WORKING DAY IF THE CURRENT DATE IS NOT A WORKING DAY
        INPUTS: DATE TO BE TESTED
        OUTPUTS: DATE
        '''
        return DatesBR().add_working_days(DatesBR().sub_working_days(date_test, 1), 1)

    def nth_weekday_month(self, date_inf, date_sup, weekday_number, nth_repetition,
                          date_format='DD/MM/YYYY', number_days_week=7):
        '''
        REFERENCES: https://stackoverflow.com/questions/57082058/get-every-nth-weekday-of-month-periodic-in-pandas
        DOCSTRING: GET NTH WORKDAY OF THE MONTH
        INPUTS:
        OUTPUTS: DATE
        '''
        list_working_days_within_dates = DatesBR().list_working_days(
            date_inf, date_sup, date_format)
        return [DatesBR().add_working_days(DatesBR().sub_working_days(d, 1), 1)
                for d in list_working_days_within_dates
                if (DatesBR().week_number(d) == weekday_number
                    and d.day >= (nth_repetition * number_days_week - number_days_week)
                    and d.day <= (nth_repetition * number_days_week))]

    def delta_calendar_days(self, inf_date, sup_date):
        '''
        DOCSTRING: COUNT CALENDAR DAYS, CONSIDERING NON-WORKING DATES
        INPUTS: INFERIOR AND SUOPERIOR DATES
        OUTPUTS: INT
        '''
        return (sup_date - inf_date).days

    def add_months(self, date_datetime, num_months):
        '''
        DOCSTRING: ADD MONTHS TO DATETIME
        INPUTS: DATE (DATETIME FORMAT) AND NUMBER OF MONTHS
        OUTPUTS: DATETIME
        '''
        return date_datetime + relativedelta(months=num_months)

    def add_calendar_days(self, original_date, days_to_add):
        '''
        DOCSTRING: ADD CALENDAR DAYS
        INPUTS: ORIGINAL DATE AND DAYS TO ADD
        OUTPUTS: DATETIME
        '''
        return original_date + timedelta(days=days_to_add)

    def delta_working_hours(self, inf_time_stamp, sup_time_stamp, start_time_office_hours=8,
                            end_time_office_hours=18, start_time_lunch_hours=0,
                            end_time_lunch_hours=0, working_days=[0, 1, 2, 3, 4]):
        '''
        REFERENCES: https://pypi.org/project/businesstimedelta/
        DOCSTRING: DELTA WORKING HOURS BETWEEN TWO TIMESTAMPS
        INPUTS: INFERIOR TIME STAMP AND SUPERIOR TIME STAMP, AS YYYY-MM-DD HH:MM:SS, START
            TIME AT OFFICE (8 AS STANDARD), END TIME AT OFFICE (18 AS STANDARD), START TIME
            FOR LUNCH (0 AS STANDARD), END TIME FOR LUNCH (0 AS STANDARD), WORKING DAYS (
                LIST FROM 0, MONDAY, TO 4, FRIDAY)
        OUTPUTS: TIME (HH:MM:SS), IN DATETIME.TIME FORMAT
        '''
        # timestamp convertation to datetime
        y_inf, mt_inf, d_inf = int(inf_time_stamp.split(' ')[0].split('-')[0]), \
            int(inf_time_stamp.split(' ')[0].split('-')[1]), \
            int(inf_time_stamp.split(' ')[0].split('-')[2])
        h_inf, m_inf, s_inf = int(inf_time_stamp.split(' ')[1].split(':')[0]), \
            int(inf_time_stamp.split(' ')[1].split(':')[1]), \
            int(inf_time_stamp.split(' ')[1].split(':')[2])
        y_sup, mt_sup, d_sup = int(sup_time_stamp.split(' ')[0].split('-')[0]), \
            int(sup_time_stamp.split(' ')[0].split('-')[1]), \
            int(sup_time_stamp.split(' ')[0].split('-')[2])
        h_sup, m_sup, s_sup = int(sup_time_stamp.split(' ')[1].split(':')[0]), \
            int(sup_time_stamp.split(' ')[1].split(':')[1]), \
            int(sup_time_stamp.split(' ')[1].split(':')[2])
        inf_time_stamp = datetime(y_inf, mt_inf, d_inf, h_inf, m_inf, s_inf)
        sup_time_stamp = datetime(y_sup, mt_sup, d_sup, h_sup, m_sup, s_sup)
        # dict of holidays
        dict_holidays_raw = dict()
        for y in range(inf_time_stamp.year, sup_time_stamp.year + 1):
            dict_holidays_raw[y] = DatesBR().holidays(y)
        dict_holidays_trt = dict()
        for k, v in dict_holidays_raw.items():
            for t in v:
                dict_holidays_trt[t[0]] = t[1]
        # office hours for working days
        workday = businesstimedelta.WorkDayRule(
            start_time=time(start_time_office_hours),
            end_time=time(end_time_office_hours),
            working_days=working_days)
        lunchbreak = businesstimedelta.LunchTimeRule(
            start_time=time(start_time_lunch_hours),
            end_time=time(end_time_lunch_hours),
            working_days=working_days)
        holidays = businesstimedelta.HolidayRule(dict_holidays_trt)
        businesshrs = businesstimedelta.Rules(
            [workday, lunchbreak, holidays])
        # output
        return businesshrs.difference(inf_time_stamp, sup_time_stamp).timedelta

    def list_last_days_of_years(self, list_years):
        '''
        DOCSTRING: LAST DAYS OF YEAR FOR A GIVEN LIST
        INPUTS: LIST OF YEARS
        OUTPUTS: LIST
        '''
        list_last_day_week_year = list()
        for y in list_years:
            last_day = date(y, 12, 31)
            if last_day.weekday() == SAT:
                list_last_day_week_year.append(last_day - timedelta(days=1))
            elif last_day.weekday() == SUN:
                list_last_day_week_year.append(last_day - timedelta(days=2))
            else:
                list_last_day_week_year.append(last_day)
        return list_last_day_week_year

    def add_holidays_not_considered_anbima(self, inf_date, sup_date, list_last_week_year_day,
                                           local_zone='pt-BR',
                                           list_holidays_not_considered=[
                                               '25/01'],
                                           list_dates_not_considered=[
                                               '05/03/2025', '18/02/2026'],
                                           list_non_bzdays_week=['sábado', 'domingo']):
        '''
        DOCSTRING: HOLIDAYS FOR SAO PAULO CALENDAR, FOR INSTANCE, THAT ARE CONSIDERED WORKING
            DAYS FOR ANBIMA OUGHT BE INCLUDED PRINCING-WISE
        INPUTS: INFERIOR DATE, SUPERIOR DATE, LIST OF LAST WEEK YEAR DAY,
            LOCAL ZONE (BRAZIL AS DEFAULT), LIST HOLIDAYS NOT CONSIDERED (AS STR DATES),
            LIST DATES NOT CONSIDERED (AS STR), LIST NON-BUSINESS DAYS
        OUTPUTS: NUMBER OF WORKING DAYS TO ADD
        '''
        locale.setlocale(locale.LC_TIME, local_zone)
        if str(type(inf_date)) == "<class 'str'>":
            inf_date = DatesBR().str_date_to_datetime(inf_date, 'DD/MM/YYYY')
        if str(type(sup_date)) == "<class 'str'>":
            sup_date = DatesBR().str_date_to_datetime(sup_date, 'DD/MM/YYYY')
        return len([d for d in DatesBR().list_calendar_days(inf_date, sup_date) if
                    (d.strftime('%d/%m') in list_holidays_not_considered and
                     not DatesBR().week_name(d) in list_non_bzdays_week or
                     d in list_last_week_year_day or d.strftime('%d/%m/%Y') in
                     list_dates_not_considered)])

    def timestamp_float_to_datetime(self, timestamp, format_timestamp='%Y-%m-%d %H:%M:%S',
                                    bl_format=True):
        '''
        DOCSTRING: TIMESTAMP FLOAT TO DATETIME
        INPUTS: TIMESTAMP AND ITS OUTPUT FORMAT AS DATETIME
        OUTPUTS: DATETIME
        '''
        if bl_format == True:
            return datetime.fromtimestamp(timestamp).strftime(format_timestamp)
        else:
            return datetime.fromtimestamp(timestamp)

    def timestamp_separator_string_to_datetime(self, str_timestamp, substring_datetime='T',
                                               format='YYYY-MM-DD'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS
        '''
        return self.str_date_to_datetime(
            StrHandler().get_string_until_substr(str_timestamp, substring_datetime), format)

    def timestamp_to_datetime(self, timestamp, bl_return_from_utc=False):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        if bl_return_from_utc == True:
            return pd.to_datetime(timestamp, unit='s', utc=True).tz_convert('America/Sao_Paulo')
        else:
            return pd.to_datetime(timestamp, unit='s', utc=True).strftime('%Y%m%d')

    def datetime_timestamp(self, date, bl_date_to_datetime=False):
        '''
        DOCSTRING: DATETIME TO TIMESTAMP
        INPUTS: DATE
        OUTPUTS: TIMESTAMP
        '''
        # converting date to datetime
        if bl_date_to_datetime == True:
            date = datetime.combine(date, datetime.min.time())
        return datetime.timestamp(date)

    def current_timestamp_string(self, format='%Y%m%d_%H%M%S'):
        '''
        DOCSTRING: CURRENT TIMESTAMP IN STRING FORMAT
        INPUTS: FORMAT (DEFAULT VALUE)
        OUTPUTS: STRING
        '''
        return DatesBR().curr_date_time().strftime(format)

    @property
    def utc_log_ts(self):
        '''
        DOCSTRING: NOW UTC LOG TIMESTAMP
        INPUTS: -
        OUTPUTS: TIMESTAMP
        '''
        return datetime.now(timezone.utc)

    def utc_from_timestamp(self, date_):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return datetime.utcfromtimestamp(date_)

    def utc_from_dt(self, date_):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        date_ = datetime.combine(date_, datetime.min.time())
        return pytz.utc.localize(date_)

    def month_year_string(self, str_dt, format_input='%b/%Y', format_output='%Y-%m', bl_dtbr=True):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        if bl_dtbr == True:
            month_mapping = {
                "JAN": "01",
                "FEB": "02",
                "MAR": "03",
                "APR": "04",
                "MAY": "05",
                "JUN": "06",
                "JUL": "07",
                "AUG": "08",
                "SEP": "09",
                "OCT": "10",
                "NOV": "11",
                "DEC": "12",
            }
            month_abbr, year = str_dt.split('/')
            month = month_mapping[month_abbr.upper()]
            return f'{year}-{month}'
        else:
            return datetime.strptime(str_dt, format_input).strftime(format_output)


### GENERAL FUNCTIONS FOR PLOTS

import matplotlib.pyplot as plt
from pathlib import Path


class GeneralPltFuncs:

    def save_fig(fig_id, tight_layout=True, fig_extension="png", resolution=300, 
                 img_dir='images', sub_dir='end_to_end_project'):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # creating directories
        images_path = Path() / img_dir / sub_dir
        images_path.mkdir(parents=True, exist_ok=True)
        # creating plot complente path name with layout of interest
        path = images_path / f"{fig_id}.{fig_extension}"
        if tight_layout:
            plt.tight_layout()
        # saving figure
        plt.savefig(path, format=fig_extension, dpi=resolution)